//--------------------------------------------------------------------------------
// Real expression class
//    Copyright (c) kei mesuda 2007
//--------------------------------------------------------------------------------

private import std.stdio;
private import std.string;
private import std.regexp;
private import std.math;
private import util.simpleparser;
private import util.rand;




class CExpressionReal : public CSimpleParser
{
public:
    static Valiable[char[]] mapValiable;
    static Rand             rand;
    static OPINFO           opOpenBrk;
    static LITINFO          litNull;

    //--------------------------------------------------
    // Execution method
    //--------------------------------------------------
    enum {
        CANNOT_EXECUTE,     // impossible to execute    
        EXECUTE_ONCE,       // execute once
        EXECUTE_MULTI       // execute multiply
    }

    //--------------------------------------------------
    // Valiable class
    //--------------------------------------------------
    static class Valiable : LITINFO
    {
    public:
        real x;
        bool isConst;
        
        this()
        {
            x = 0;
            isConst = false;
        }
        
        this(real x_)
        {
            x = x_;
            isConst = true;
        }
        
        int calc(Branch br)
        {
            stac(x);
            return 0;
        }
        
        int assign(Branch br)
        {
            if (isConst) return 1;
            x = stac[0][0];
            return 0;
        }
    }


public:
    //--------------------------------------------------
    // Stack area class
    //--------------------------------------------------
    static class STACKAREA
    {
    public:
        real   initValue;   // initial value
        real[] s;           // stacking area
        int    i;           // surface index
        
    public:
        this(real initval)  { s=new real[8]; reset(); initValue=initval; }
        // stac() sets surface value
        void opCall(real x) { s[i] = x; }
        // stac[] returns slice[1] indexed from surface
        real[]  opIndex(uint idx) { assert(i>=idx); return s[i-idx..i-idx+1]; }
        
        void   reset()     { i=0; s[0]=initValue; }
        int    push()      { if (s.length<=++i) { s.length = s.length << 1; } s[i] = initValue; return i-1; }
        void   pop()       { assert(i>0); --i; }
        real[] pop(int i_) { int end=i; i=i_; return s[i+1..end+1]; }
    }

    // Stack area for calculation
    static STACKAREA        stac;


public:
    //--------------------------------------------------
    // Monomial operator information class
    //--------------------------------------------------
    static class MOPINFO : OPINFO
    {
    public:
        // caluclating function
        void function(inout real) fn;

        // constructor
        this(char[] key_, byte prior, void function(inout real) fn_)
        {
            super(1, key_, prior);
            fn = fn_;
        }
        
        // Is recordable ?
        bool isRecordable() { return (fn !is null); }
        
        // calculation
        int calc(Branch br)
        {
            int jump_idx = br.br0.calc();   // calculate 1st branch
            fn(stac[0][0]);                 // call function
            return jump_idx;
        }
    }


    //--------------------------------------------------
    // Monomial assigning operator information class
    //--------------------------------------------------
    static class MAOPINFO : OPINFO
    {
    public:
        // caluclating function
        void function(out real, inout real) fn;

        // constructor
        this(char[] key_, byte prior, void function(out real, inout real) fn_)
        {
            super(1, key_, prior);
            fn = fn_;
        }

        // calculation
        int calc(Branch br)
        {
            br.br0.calc();                  // calculate 1st branch
            stac.push();                    // push stack
            fn(stac[0][0], stac[1][0]);     // call function
            int jump_idx = br.br0.assign(); // assign 1st branch
            stac.pop();                     // pop stack
            return jump_idx;
        }
    }

    
    //--------------------------------------------------
    // Binomial operator information class
    //--------------------------------------------------
    static class BOPINFO : OPINFO
    {
    public:
        void   function(inout real, real) fn;       // caluclating function
        ubyte  function(inout int)        selector; // 2nd operand selector

        // constructor
        this(char[] key_, byte prior, void function(inout real, real) fn_, ubyte function(inout int) selector_)
        {
            super(2, key_, prior);
            fn = fn_;
            selector = (selector_ != null) ? (selector_) : &(selectorDefault);
        }
        
        // calculation
        int calc(Branch br)
        {
            int  jump_idx;
            byte exec_status;
            
            do {
                jump_idx = br.br0.calc();               // calculate 1st operand
                exec_status = selector(jump_idx);       // select execution method
                if (exec_status != CANNOT_EXECUTE) {    // calculate 2nd operand only when selector return "EXECUTE_*".
                    stac.push();                        // push stack
                    jump_idx = br.br1.calc();           // calculate 2nd branch
                    fn(stac[1][0], stac[0][0]);         // call function
                    stac.pop();                         // pop stack
                }
            } while (exec_status == EXECUTE_MULTI);     // multiply execution
            
            return jump_idx;
        }
    }

    
    //--------------------------------------------------
    // Binomial assigning operator information class
    //--------------------------------------------------
    static class BAOPINFO : OPINFO
    {
    public:
        void   function(out real, inout real, real) fn;        // caluclating function
        ubyte  function(inout int)                  selector;  // 2nd operand selector

        // constructor
        this(char[] key_, byte prior, void function(out real, inout real, real) fn_, ubyte function(inout int) selector_)
        {
            super(2, key_, prior);
            fn = fn_;
            selector = (selector_ != null) ? (selector_) : &(selectorDefault);
        }

        // calculation
        int calc(Branch br)
        {
            int  jump_idx;
            byte exec_status;
            
            do {
                jump_idx = br.br0.calc();                   // calculate 1st operand
                exec_status = selector(jump_idx);           // select execution method
                if (exec_status != CANNOT_EXECUTE) {        // calculate 2nd operand only when selector return "EXECUTE_*".
                    stac.push();                            // push stack
                    stac.push();                            // push stack
                    jump_idx = br.br1.calc();               // calculate 2nd branch
                    fn(stac[1][0], stac[2][0], stac[0][0]); // call function
                    stac.pop();                             // pop stack
                    jump_idx = br.br1.assign();             // assign 2nd branch
                    stac.pop();                             // pop stack
                }
            } while (exec_status == EXECUTE_MULTI);         // multiply execution
            
            return jump_idx;

        }
    }


    //--------------------------------------------------
    // function operator information class
    //--------------------------------------------------
    static class FUNCINFO : OPINFO
    {
    public:
        // caluclating function
        real function(real[])[16] fn;

        // constructor
        this(char[] key_, byte prior)
        {
            super(1, key_, prior);
        }
        
        // overload
        FUNCINFO overload(int param_count, real function(real[]) fn_)
        {
            assert(param_count<16);
            fn[param_count] = fn_;
            return this;
        }
        
        // calculation
        int calc(Branch br)
        {
            if (br.br0.info is litNull) {
                if (fn[0] != null) { stac(fn[0](null)); return 0; }
            } else {
                int stac_idx = stac.push();         // push stack
                int jump_idx = br.br0.calc();       // calculate 1st branch
                real[] param = stac.pop(stac_idx);  // pop stack
                assert(param.length < 16);
                if (fn[param.length] != null) {
                    stac(fn[param.length](param));  // call function
                    return 0;
                }
            }
            return 1;
        }
    }


    
    
    // Resister bracket
    static OPINFO addBracket(char[] keyOpen, char[] keyClose)
    {
        OPINFO opClose = addMonomialOperator(new MOPINFO(keyClose,  1, null));
        OPINFO opOpen  = addMonomialOperator(new MOPINFO(keyOpen,  -1, null));
        opOpen.opRelative = opClose;
        return opOpen;
    }

    
    // Register monomial operator 
    static OPINFO addOperator(char[] key, void function(inout real) fn, byte prior)
    {
        return addMonomialOperator(new MOPINFO(key, prior, fn));
    }


    // Register monomial assigning operator 
    static OPINFO addOperator(char[] key, void function(out real, inout real) fn, byte prior)
    {
        return addMonomialOperator(new MAOPINFO(key, prior, fn));
    }


    // Register binomial operator
    static OPINFO addOperator(char[] key, void function(inout real, real) fn, byte prior, 
                              ubyte function(inout int) selector = null)
    {
        return addBinomialOperator(new BOPINFO(key, prior, fn, selector));
    }


    // Register binomial assign operator
    static OPINFO addOperator(char[] key, void function(out real, inout real, real) fn, byte prior, 
                              ubyte function(inout int) selector = null)
    {
        return addBinomialOperator(new BAOPINFO(key, prior, fn, selector));
    }

    
    // Register function
    static FUNCINFO addFunction(char[] key, byte prior)
    {
        FUNCINFO newFunc = new FUNCINFO(key, prior);
        addMonomialOperator(newFunc);
        return newFunc;
    }

    // add contant value
    static void addValue(char[] str, real x)
    {
        mapValiable[str] = new Valiable(x);
    }
    



// Virtual functions
public:
    // Parser for real literal
    bool parseLiteral(out LITINFO lit, inout char[] str)
    {
        char[] tok;

        // literal
        scope RegExp reglit = RegExp(r"^[+-]?[0-9]+\.?[0-9]*(e[+-]?[0-9]+)?");
        if (reglit.find(str) != -1) {
            tok = reglit.match(0);
            str = str[tok.length..length];
            lit = new Valiable(atof(tok));
            
            debug{ writefln(tok); }
            return true;
        }
        
        // valiable
        scope RegExp regval = RegExp(r"^([_A-Za-z$][_A-Za-z0-9]*)");
        if (regval.find(str) != -1) {
            tok = regval.match(1);
            str = str[regval.match(0).length..length];
            
            Valiable* plit = (tok in mapValiable);
            if (plit != null) {
                lit = *plit;
            } else {
                mapValiable[tok] = new Valiable();
                lit = mapValiable[tok];
            }
            
            debug{ writefln(tok); }
            return true;
        }
        
        // null... ('A`)
        char[] strClose = opOpenBrk.opRelative.key;
        if (o_stac.back() is opOpenBrk && str[0..strClose.length]==strClose) {
            lit = litNull;
            return true;
        }
        
        return false;
    }
    



// 2nd operand execution selector
private:
    // Default 2nd operand execution selector
    static ubyte selectorDefault(inout int jump_idx)
    {
        jump_idx = 0;
        return EXECUTE_ONCE;
    }

    // 2nd operand selector for jump function
    //  stac[0] holds 1st operand's answer
    static ubyte selectorTrueExecute(inout int jump_idx)
    {
        // if 1st operand's answer != 0, jump_idx = 0
        jump_idx = (stac[0][0] != 0) ? 0 : 1;
        // when jump_idx = 0, EXECUTE_ONCE
        return (jump_idx==0) ? EXECUTE_ONCE : CANNOT_EXECUTE;
    }
    
    static ubyte selectorFalseExecute(inout int jump_idx)
    {
        // if 1st operand's answer == 0, jump_idx = 0
        jump_idx = (stac[0][0] == 0) ? 0 : 1;
        // when jump_idx = 0, EXECUTE_ONCE
        return (jump_idx==0) ? EXECUTE_ONCE : CANNOT_EXECUTE;
    }
    
    static ubyte selectorTrueLoop(inout int jump_idx)
    {
        // if 1st operand's answer != 0, jump_idx = 0
        jump_idx = (stac[0][0] != 0) ? 0 : 1;
        // when jump_idx = 0, EXECUTE_ONCE
        return (jump_idx==0) ? EXECUTE_MULTI : CANNOT_EXECUTE;
    }
    
    static ubyte selectorFalseLoop(inout int jump_idx)
    {
        // if 1st operand's answer == 0, jump_idx = 0
        jump_idx = (stac[0][0] == 0) ? 0 : 1;
        // when jump_idx = 0, EXECUTE_ONCE
        return (jump_idx==0) ? EXECUTE_MULTI : CANNOT_EXECUTE;
    }
    
    static ubyte selectorIndexJump(inout int jump_idx)
    {
        // 1st operand's answer directly translate to jump_idx.
        jump_idx = (stac[0][0] < 0 || int.max <= stac[0][0]) ? int.max : cast(int)(stac[0][0]);
        return (jump_idx == 0) ? EXECUTE_ONCE : CANNOT_EXECUTE;
    }

    static ubyte selectorIndexCatch(inout int jump_idx)
    {
        // Decrease jump_idx, and execute when jump_idx==0.
        --jump_idx;
        return (jump_idx == 0) ? EXECUTE_ONCE : CANNOT_EXECUTE;
    }

    static ubyte selectorIndexFinalCatch(inout int jump_idx)
    {
        // Execute when jump_idx > 0
        ubyte result = (jump_idx > 0) ? EXECUTE_ONCE : CANNOT_EXECUTE;
        // jump_idx = 0 means no error
        jump_idx = 0;
        return result;
    }


    

// operation
public:
    // Function for binomial operator
    //  ans holds 1st operand's answer, set the function answer.
    //  src holds 2nd operands' answer.
    //  set the assigning value to asg.
    static void fnThrough(inout real ans, real src) { ans = src; }
    static void fnStac(inout real ans, real src) { stac.push(); }
    static void fnAdd(inout real ans, real src) { ans += src; }
    static void fnSub(inout real ans, real src) { ans -= src; }
    static void fnMul(inout real ans, real src) { ans *= src; }
    static void fnDiv(inout real ans, real src) { ans /= src; }
    static void fnSup(inout real ans, real src) { ans -= trunc(ans / src) * src; }  // ?? modf ??
    static void fnPow(inout real ans, real src) { ans = pow(ans, src); }
    
    static void fnAssign(out real asg, inout real ans, real src) { asg = ans; }
    static void fnAddasg(out real asg, inout real ans, real src) { asg = ans = src + ans; }
    static void fnSubasg(out real asg, inout real ans, real src) { asg = ans = src - ans; }
    static void fnMulasg(out real asg, inout real ans, real src) { asg = ans = src * ans; }
    static void fnDivasg(out real asg, inout real ans, real src) { asg = ans = src / ans; }
    static void fnSupasg(out real asg, inout real ans, real src) { asg = ans = src - trunc(src / ans) * (ans); } // ?? modf ??
    static void fnPowasg(out real asg, inout real ans, real src) { asg = ans = pow(src, ans); }

    static void fnEql(inout real ans, real src)    { ans = (ans == src) ? 1 : 0; }
    static void fnNeq(inout real ans, real src)    { ans = (ans != src) ? 1 : 0; }
    static void fnSeq(inout real ans, real src)    { ans = (ans <= src) ? 1 : 0; }
    static void fnLeq(inout real ans, real src)    { ans = (ans >= src) ? 1 : 0; }
    static void fnSml(inout real ans, real src)    { ans = (ans <  src) ? 1 : 0; }
    static void fnLrg(inout real ans, real src)    { ans = (ans >  src) ? 1 : 0; }
    static void fnXorXor(inout real ans, real src) { ans = ((ans != 0 && src == 0) || (ans == 0 && src != 0)) ? 1 : 0; }

    // Function for monomial operator
    //  ans holds 1st operand's answer, set the function answer.
    //  set the assigning value to asg.
    static void fnInc   (out real asg, inout real ans) { asg = (ans += 1); }
    static void fnDec   (out real asg, inout real ans) { asg = (ans -= 1); }
    static void fnIncAft(out real asg, inout real ans) { asg = ans + 1; }
    static void fnDecAft(out real asg, inout real ans) { asg = ans - 1; }
    static void fnNop   (inout real ans) { }
    static void fnNeg   (inout real ans) { ans = -ans; }
    static void fnOpp   (inout real ans) { ans = (ans != 0) ? 0 : 1; }
    static void fnInt   (inout real ans) { ans = trunc(ans); }
    static void fnCeil  (inout real ans) { ans = ceil(ans);  }
    static void fnFloor (inout real ans) { ans = floor(ans); }
    static void fnRound (inout real ans) { ans = (ans >= 0) ? (floor(ans + 0.5)) : (ceil(ans - 0.5)); }
    static void fnACeil (inout real ans) { ans = (ans >= 0) ? (ceil(ans))        : (floor(ans)); }
    static void fnAbs   (inout real ans) { ans = abs(ans); }
    static void fnSign  (inout real ans) { ans = (ans > 0) ? 1 : ((ans < 0) ? -1 : 0); }
    static void fnBool  (inout real ans) { ans = (ans != 0) ? 1 : 0; }
    static void fnDegree(inout real ans) { ans = ans * 0.017453292519943295769236907684886; }
    static void fnMulPi (inout real ans) { ans = ans * PI; }
    
    // Function
    //  prm holds parameter, and return the answer.
    static real fnSin  (real[] prm)  { return sin(prm[0]); }
    static real fnCos  (real[] prm)  { return cos(prm[0]); }
    static real fnTan  (real[] prm)  { return tan(prm[0]); }
    static real fnAsn  (real[] prm)  { return asin(prm[0]); }
    static real fnAcs  (real[] prm)  { return acos(prm[0]); }
    static real fnAtn  (real[] prm)  { return atan(prm[0]); }
    static real fnAtn2 (real[] prm)  { return atan2(prm[1], prm[0]); }
    static real fnSinh (real[] prm)  { return sinh(prm[0]); }
    static real fnCosh (real[] prm)  { return cosh(prm[0]); }
    static real fnTanh (real[] prm)  { return tanh(prm[0]); }
    static real fnAsnh (real[] prm)  { return asinh(prm[0]); }
    static real fnAcsh (real[] prm)  { return acosh(prm[0]); }
    static real fnAtnh (real[] prm)  { return atanh(prm[0]); }
    static real fnLoge (real[] prm)  { return log(prm[0]); }
    static real fnLog  (real[] prm)  { return log2(prm[0]) / log2(prm[1]); }
    static real fnLog2 (real[] prm)  { return log2(prm[0]); }    
    static real fnLog10(real[] prm)  { return log10(prm[0]); }
    static real fnRnd0 (real[] prm)  { return rand.getr(1); }
    static real fnRnd1 (real[] prm)  { return rand.getr(prm[0]); }
    static real fnRnd2 (real[] prm)  { return rand.getr(prm[0]-prm[1])+prm[1]; }
    static real fnRnds0(real[] prm)  { return rand.getsr(1); }
    static real fnRnds1(real[] prm)  { return rand.getsr(prm[0]); }
    static real fnRnds2(real[] prm)  { 
        real rng = prm[0]-prm[1];
        real rnd = rand.getr(rng);
        return (rnd<rng*0.5) ? (-rnd-prm[1]) : (rnd+prm[1]);
    }




    // make rule
    static this()
    {
        // Initialize
        stac = new STACKAREA(0);
        rand = new Rand();
        litNull = new LITINFO();

        // Binomial
        addOperator("=",  &fnAssign, -3);
        addOperator("+=", &fnAddasg, -3);
        addOperator("-=", &fnSubasg, -3);
        addOperator("**=",&fnPowasg, -3);
        addOperator("*=", &fnMulasg, -3);
        addOperator("/=", &fnDivasg, -3);
        addOperator("%=", &fnSupasg, -3);
        
        addOperator("+",  &fnAdd, 12);
        addOperator("-",  &fnSub, 12);
        addOperator("**", &fnPow, 14);
        addOperator("*",  &fnMul, 13);
        addOperator("/",  &fnDiv, 13);
        addOperator("%",  &fnSup, 13);
        
        addOperator("<=", &fnSeq,  9);
        addOperator(">=", &fnLeq,  9);
        addOperator("<",  &fnSml,  9);
        addOperator(">",  &fnLrg,  9);
        addOperator("==", &fnEql,  8);
        addOperator("!=", &fnNeq,  8);
        
        addOperator("||", &fnThrough, 5, &selectorFalseExecute);     // Only if stac[1]==0, fnThrough() is called.
        addOperator("&&", &fnThrough, 6, &selectorTrueExecute);      // Only if stac[1]!=0, fnThrough() is called.
        addOperator("^^", &fnXorXor,  5);
        
        addOperator("?i", &fnThrough, 4, &selectorIndexJump);        // stac[1] translate to jump index.
        addOperator("?w", &fnThrough, 4, &selectorTrueLoop);         // Only if stac[1]!=0, fnThrough() is called. :=while
        addOperator("?u", &fnThrough, 4, &selectorFalseLoop);        // Only if stac[1]==0, fnThrough() is called. :=until
        addOperator("?",  &fnThrough, 4, &selectorTrueExecute);      // Only if stac[1]!=0, fnThrough() is called. :=if
        addOperator(":<", &fnThrough, 4, &selectorIndexFinalCatch);  // Final catcher. := else
        addOperator(":",  &fnThrough, 4, &selectorIndexCatch);       // If operand index is jumped, fnThrough() is called.
        
        addOperator(",",  &fnStac,   -2);

        // Prefix type monomial
        addOperator("++", &fnInc, -64);   // increase before
        addOperator("--", &fnDec, -64);   // decrease before
        addOperator("+",  &fnNop, -64);   // no operation for prefixed "+"
        addOperator("-",  &fnNeg, -64);   // negative
        addOperator("!",  &fnOpp, -64);   // bool negative
        addOperator("bool",  &fnBool,  -64);   // bool;   x!=0 changed into 1. x=0 keeped.
        addOperator("int",   &fnInt,   -64);   // integer;rounded approcimate to 0.     [-] -> 0 <- [+]
        addOperator("expand",&fnACeil, -64);   // expand; rounded away form 0.          [-] <- 0 -> [+]
        addOperator("floor", &fnFloor, -64);   // floor;  rounded downward toward -inf. [-] <- 0 <- [+]
        addOperator("ceil",  &fnCeil,  -64);   // ceil;   rounded upward   toward +inf. [-] -> 0 -> [+]
        addOperator("round", &fnRound, -64);   // round;  rounded nearest integer. Value of +-0.5 rounded away form 0
        addOperator("abs",   &fnAbs,   -64);   // abs;    absolution
        addOperator("sign",  &fnSign,  -64);   // sign;   x<0 changed into -1, x>0 changed into 1. x=0 keeped.
        
        // Postfix type monomial
        addOperator("++", &fnIncAft, 63); // increase after
        addOperator("--", &fnDecAft, 63); // decrease after
        addOperator("lf", &fnFloor,  63); // floor;  rounded downward toward -inf. [-] <- 0 <- [+]
        addOperator("lc", &fnCeil,   63); // ceil;   rounded upward   toward +inf. [-] -> 0 -> [+]
        addOperator("lx", &fnACeil,  63); // expand; rounded away form 0.          [-] <- 0 -> [+]
        addOperator("lr", &fnRound,  63); // round;  rounded nearest integer. Value of +-0.5 rounded away form 0
        addOperator("l",  &fnInt,    63); // integer;rounded approcimate to 0.     [-] -> 0 <- [+]
        addOperator("a",  &fnAbs,    63); // abs;    absolution
        addOperator("s",  &fnSign,   63); // sign;   x<0 changed into -1, x>0 changed into 1. x=0 keeped.
        addOperator("b",  &fnBool,   63); // bool;   x!=0 changed into 1. x=0 keeped.
        addOperator("deg",&fnDegree, 63); // degree; translate degree to radian
        addOperator("pi", &fnMulPi,  63); // pi;     multiple PI
        
        // functions
        addFunction("sinh",  -64).overload(1, &fnSinh);
        addFunction("cosh",  -64).overload(1, &fnCos);
        addFunction("tanh",  -64).overload(1, &fnTanh);
        addFunction("asinh", -64).overload(1, &fnAsnh);
        addFunction("acosh", -64).overload(1, &fnAcsh);
        addFunction("atanh", -64).overload(1, &fnAtnh);
        addFunction("sin",   -64).overload(1, &fnSin);
        addFunction("cos",   -64).overload(1, &fnCos);
        addFunction("tan",   -64).overload(1, &fnTan);
        addFunction("asin",  -64).overload(1, &fnAsn);
        addFunction("acos",  -64).overload(1, &fnAcs);
        addFunction("atan",  -64).overload(1, &fnAtn).overload(2, &fnAtn2);
        addFunction("log10", -64).overload(1, &fnLog10);
        addFunction("log2",  -64).overload(1, &fnLog2);
        addFunction("loge",  -64).overload(1, &fnLoge);
        addFunction("lg",    -64).overload(1, &fnLog2);
        addFunction("ln",    -64).overload(1, &fnLoge);
        addFunction("log",   -64).overload(1, &fnLog10).overload(2, &fnLog);
        addFunction("rnds",  -64).overload(0, &fnRnds0).overload(1, &fnRnds1).overload(2, &fnRnds2);
        addFunction("rnd",   -64).overload(0, &fnRnd0) .overload(1, &fnRnd1) .overload(2, &fnRnd2);

        // Bracket
        opOpenBrk = addBracket("(", ")");


        // Constant value
        addValue("false", 0);
        addValue("true",  1);
        addValue("pi",    PI);
        addValue("exp",   E);
        addValue("inf",   real.infinity);
        addValue("nan",   real.nan);
    }


public:
    real exec()
    {
        stac.reset();
        super.reset();
        super.calc();
        return stac[0][0];
    }
}

