//--------------------------------------------------------------------------------
// Simple lexical and syntax parsing class.
//    Copyright (c) kei mesuda 2007
//--------------------------------------------------------------------------------

private import std.stdio;
private import std.string;
private import util.list;




//--------------------------------------------------------------------------------
// Simple lexical and syntax parsing class. This class is not legitimate.
//--------------------------------------------------------------------------------
class CSimpleParser
{
protected:
    //--------------------------------------------------
    // Binary branch type
    //--------------------------------------------------
    enum {
        BRANCH_TERMINAL,   // terminator
        BRANCH_LITERAL,    // literal leaf
        BRANCH_OPERATOR    // operator branch
    }


    //--------------------------------------------------
    // Binary branch information base class
    //--------------------------------------------------
    static abstract class BRINFO
    {
    public:
        ubyte  type;        // operand type
        byte   param_count; // parameter count
        
    public:
        // constructor
        this(ubyte type_, byte param_count_)
        {
            type = type_; 
            param_count = param_count_;
        }
        
        // virtual function
        int calc(Branch br)   { return 0; }  // calculation
        int assign(Branch br) { return 1; }  // assigning
    }

    
    //--------------------------------------------------
    // Literal leaf information base class
    //--------------------------------------------------
    static class LITINFO : BRINFO
    {
    public:
        this() { super(BRANCH_LITERAL, 0); }
    }
    
    
    //--------------------------------------------------
    // Operator branch information base class
    //--------------------------------------------------
    // <MEMO> about construction parameter "proir"
    //  Positive prior means left-priority, negative means right-priority.
    //  And absolute value of prior means execition priority.
    //  but -1, 0, 1 of prior have a special meaning.
    //  -1 of prior means entire right-priority; [priority left]=0,   [priority right]=max
    //   1 of prior means entire left-priority;  [priority left]=max, [priority right]=0
    //   0 of prior means entire lowest priority;[priority left]=0,   [priority right]=0
    //  Usually, binomial operator has POSITIVE, and assigning binomial has NEGATIVE value.
    //  Prefix-type monomial operator has NEGATIVE, and suffix-type monomial has POSITIVE value.
    //  Open bracket (like "(","[","{"...) has -1, close bracket has 1.
    //--------------------------------------------------
    static class OPINFO : BRINFO
    {
    public:
        char[] key;         // letters for operator
        ubyte  priorL;      // priority of left oprend
        ubyte  priorR;      // priority of right oprend
        OPINFO opRelative;  // relative operator

        // constructor
        this(byte param_count_, char[] key_, byte prior)
        {
            super(BRANCH_OPERATOR, param_count_);
            key = key_;
            _translate_priority_from_byte(prior);
        }

        // constructor for terminal operator
        this()
        {
            super(BRANCH_TERMINAL, 2);
            _translate_priority_from_byte(0);
        }
        
        // Is right prior high ?
        bool isRightProir() { return (priorL < priorR); }
        
        // Is recordable ?
        bool isRecordable() { return true; }

    private:
        // Calculate priority
        void _translate_priority_from_byte(byte pr)
        {
            switch (pr) {
            case -1:
                priorL=0;
                priorR=ubyte.max;
                return;
            case 1:
                priorL=ubyte.max;
                priorR=0;
                return;
            case 0:
                priorL=0;
                priorR=0;
                return;
            default:
                // right priority flag
                bool  rpr = (pr<0);
                // upr = |pr| * 2 + 1  
                ubyte upr = cast(ubyte) ((((rpr) ? (-pr) : (pr)) << 1) + 1);
                priorL = upr + (!rpr);
                priorR = upr + (rpr);
            }
        }
    }


    //--------------------------------------------------
    // Binary tree element base class
    //--------------------------------------------------
    class Branch
    {
    public:
        // branching infomation
        BRINFO info;           // information for this branch
        Branch br0 = null;     // first branch
        Branch br1 = null;     // second branch
        Branch parent = null;  // parent branch

        // executing infomation
        bool   is_calculated;  // this flag means that this branch is already calculated
        
        // contract
        invariant {
            switch(info.param_count) {
            case 0:  assert(br0  is null && br1  is null);    break;
            case 1:  assert(br0 !is null && br1  is null);    break;
            case 2:  assert(br0 !is null && br1 !is null);    break;
            default: assert(false);
            }
        }
        
        // constructor for leaf
        this(BRINFO info_)
        {
            info = info_;
        }
        
        // constructor for branch
        this(BRINFO info_, Branch br0_, Branch br1_)
        {
            info = info_;
            br0  = br0_;
            br1  = br1_;
            if (br0 !is null) br0.parent = this;
            if (br1 !is null) br1.parent = this;
        }

        // execution
        ubyte type()   { return info.type; }
        int   calc()   { int result=info.calc(this);   is_calculated = true; return result; }
        int   assign() { int result=info.assign(this); is_calculated = true; return result; }
        
        // reset
        void reset()
        {
            if (br0 !is null) br0.reset();
            if (br1 !is null) br1.reset();
            is_calculated = false;
        }
    }

        


//--------------------------------------------------------------------------------
// Static members
//--------------------------------------------------------------------------------
protected:
    static list!(OPINFO) listPrefix;     // list for prefix operator
    static list!(OPINFO) listSuffix;     // list for suffix operator
    static list!(OPINFO) listBinomial;   // list for binomial operator
        
    // Initial value
    static OPINFO opTerminal;

    // Stack area for parsing
    static list!(OPINFO)  o_stac;  // opertor information stack
    static list!(Branch)  b_stac;  // branch stack

    // constructor for statics
    static this()
    {
        listPrefix   = new list!(OPINFO);
        listSuffix   = new list!(OPINFO);
        listBinomial = new list!(OPINFO);
        
        o_stac       = new list!(OPINFO);
        b_stac       = new list!(Branch);
        
        opTerminal   = new OPINFO();
    }

    


//--------------------------------------------------------------------------------
// Member values
//--------------------------------------------------------------------------------
protected:
    // root of binary tree
    Branch    m_Root;
    
    this() { }




//--------------------------------------------------
// Create parser rule
//--------------------------------------------------
public:
    // Register monomial operator
    static OPINFO addMonomialOperator(OPINFO op)
    {
        if (op.isRightProir()) listPrefix ~= op;
        else                   listSuffix ~= op;
        return op;
    }


    // Register binomial operator
    static OPINFO addBinomialOperator(OPINFO op)
    {
        listBinomial ~= op;
        return op;
    }


// Virtual functions
public:
    // Branch creater for literal and operator
    Branch newBranch(LITINFO lit)                        { return new Branch(lit); }
    Branch newBranch(OPINFO  op, Branch br0, Branch br1) { return new Branch(op, br0, br1); }

    // Literal parser
    bool parseLiteral(out LITINFO outlit, inout char[] str)
    {
        return false;
    }

    // Operator parser
    bool parseOperator(list!(OPINFO) oplist, out OPINFO outop, inout char[] str)
    {
        // Return terminator when parsing is done.
        if (str.length == 0) {
            outop = opTerminal;
            return true;
        }

        // Search operator
        foreach (OPINFO op; oplist) {
            if (op.key.length <= str.length && str[0..op.key.length] == op.key) {
                str = str[op.key.length..length];
                outop = op;
                return true;
            }
        }

        // Nothing to match
        return false;
    }




//--------------------------------------------------
// parse() sub routines. 
//  Return value; -1 to error or terminated.
//                 0 to recall this.
//                 1 to go next.
//                >2 to skip next.
//--------------------------------------------------
private:
    int _parsesubLiteral(inout char[] str, inout bool isMatched)
    {
        // Remove white space
        str = stripl(str);

        // Parse literal
        LITINFO lit;
        if (parseLiteral(lit, str)) {
            isMatched |= true;
            b_stac.push_back(newBranch(lit));
            // go next
            return 1;
        }

        // error
        return -1;
    }


    // next_idx means recall or go next when success to parse.
    int _parsesubOperator(list!(OPINFO) oplist, int next_idx, inout char[] str, inout bool isMatched)
    {
        // Remove next white space
        str = stripl(str);

        // Parse prefix and suffix monomial and open bracket operator
        OPINFO op;
        if (parseOperator(oplist, op, str)) {
            isMatched |= true;
            if (!_addOperatorBranch(op)) return -1;     // !! Failure !!
            debug { writefln(op.key); }
            return next_idx;                            // recall or go next
        }

        // go next
        return 1;
    }




//--------------------------------------------------
// parse() sub-sub routines.
//--------------------------------------------------
private:
    // Sub-routine to add operator branch
    bool _addOperatorBranch(OPINFO op)
    {
        OPINFO  op_back;
        
        while (true) {
            // When o_stac.priority < op.priority, stack op to o_stac.
            if (o_stac.back().priorL < op.priorR) {
                o_stac.push_back(op);
                return true;
            }

            // When o_stac.priority >= op.priority, ...
            op_back = o_stac.pop_back();
            switch (op_back.type) {
            case BRANCH_OPERATOR:
                if (op_back.opRelative !is null && 
                    op_back.opRelative !is op)  return false;   // check relation
                if (!op_back.isRecordable())    return true;    // do nothing when op_back.isRecordable()==false
                if (!_compressOperand(op_back)) return false;   // create new operator branch
                break;                                          // if compression successes, go next loop
            case BRANCH_TERMINAL:
                return false;                                   // falure or finish
            default:    // BRANCH_LITERAL
                assert(false);                                  // ('A`)
            }
        }
        
        // never execute
        return false;
    }

    
    // Create new operand from b_stac surface, and push_back b_stac.
    bool _compressOperand(OPINFO op)
    {
        // Check b_stac size
        if (b_stac.size() < op.param_count) return false;

        Branch brOperator;
        Branch br0, br1;
        
        switch (op.param_count) {
        case 2:
            // binomial operator
            if (op.isRightProir()) {
                // In right-priority operator, b_stac.back() is the 1st branch.
                br0 = b_stac.pop_back();
                br1 = b_stac.pop_back();
            } else {
                // In left-priority operator, b_stac.back() is the 2nd branch.
                br1 = b_stac.pop_back();
                br0 = b_stac.pop_back();
            }
            brOperator = newBranch(op, br0, br1);
            break;
        case 1:
            // monomial operator
            brOperator = newBranch(op, b_stac.pop_back(), null);
            break;
        default:
            // literal (no operand) operator
            brOperator = newBranch(op, null, null);
            break;
        }
        
        // compressing
        b_stac.push_back(brOperator);

        return true;
    }


    // Check termination when _addOperatorBranch() returns false.
    bool _checkTermination(char[] str)
    {
        if (str.length == 0 && b_stac.size() == 1 && o_stac.size() == 0) {
            m_Root = b_stac.pop_back();
            return true;
        }
        return false;
    }

    


//--------------------------------------------------
// public functions 
//--------------------------------------------------
public:
    // lexical and syntax analysis
    bool parse(char[] str)
    {
        // temporals
        bool    isMatched;
        int     iStatus;
        
        // initialize
        m_Root = null;
        // Insert terminal operator at first.
        b_stac.erase();
        o_stac.erase();
        o_stac.push_back(opTerminal);
        
        // parsing
        iStatus = 0;
        do {
            isMatched = false;
            while (iStatus == 0)  { iStatus = _parsesubOperator(listPrefix, 0, str, isMatched); }   // prefix monomial
            if    (iStatus == -1) { return false; }
            --iStatus;
            while (iStatus == 0)  { iStatus = _parsesubLiteral(str, isMatched); }                   // literal
            if    (iStatus == -1) { return false; }
            --iStatus;
            while (iStatus == 0)  { iStatus = _parsesubOperator(listSuffix, 0, str, isMatched); }   // suffix monomial
            if    (iStatus == -1) { return (_checkTermination(str)); }                              // success if termineted here
            --iStatus;
            while (iStatus == 0)  { iStatus = _parsesubOperator(listBinomial, 1, str, isMatched); } // binomial
            if    (iStatus == -1) { return false; }
            --iStatus;
        } while (isMatched);
        
        // no matching
        return false;
    }

        
    // reset
    void reset()
    {
        m_Root.reset();
    }

    
    // calculation
    void calc()
    {
        m_Root.calc();
    }


    // get next execution branch
    Branch get_next(Branch br)
    {
        if (br.br0 !is null && !br.br0.is_calculated) return get_next(br.br0);
        if (br.br1 !is null && !br.br1.is_calculated) return get_next(br.br1);
        if (br.is_calculated) return (br.parent is null) ? null : get_next(br.parent);
        br.is_calculated = true;
        return br;
    }
}

