//---------------------------------------------------------------------------------
//  Window base class
//---------------------------------------------------------------------------------
module util_win.window;

private import win32.windows;
private import util.list;
private import util_win.global;

alias list!(IWindow)    listWindow;


public struct WINDOW_SETTING
{
    char[]  title  = "DApp";    // Window title

    HICON   hIcon  = null;      // Icon
    HMENU   hMenu  = null;      // Menu
    HACCEL  hAccel = null;      // Accel

    uint    style    = WS_VISIBLE | WS_CAPTION | WS_SYSMENU;    // Window style
    uint    styleEx  = 0;                                       // Expanded window style
    POINT   ptWindow = {x:CW_USEDEFAULT, y:CW_USEDEFAULT};      // Window position
    POINT   szClient = {x:240, y:240};                          // Window client size
    POINT   szWindow;                                           // Window size
}




//---------------------------------------------------------------------------------
//  Window base class
//---------------------------------------------------------------------------------
class IWindow
{
//---------------------------------------------------
//  Member values
//---------------------------------------------------
protected:
    WINDOW_SETTING  m_setting;      /// Setting infomation
    HWND            m_hWnd;         /// Window handle
    HWND            m_hParent;      /// Parent window handle
    listWindow      m_listChild;    /// List of child windows

    
//---------------------------------------------------
//  Member function
//---------------------------------------------------
public:
    /// Constructor
    this()
    {
        m_listChild = new listWindow;
    }
    

    /// Constructor with handle
    this(HWND hWnd)
    {
        m_hWnd = hWnd;
    }
    
    
    /// Get window handle
    HWND getHandle()
    {
        return m_hWnd;
    }
    
    
    /// Set window handle
    void formHandle(HWND hWnd)
    {
        m_hWnd = hWnd;
    }


    /// Get window setting
    WINDOW_SETTING* getSetting()
    {
        return &m_setting;
    }


    

//----------------------------------------------------------------------
//  Operation 
//----------------------------------------------------------------------
public:
    bool create(HWND hParent)
    {
        // Parent window handle
        m_hParent = hParent;
        
        // Calcurate certain window size
        RECT  rc = calcWindowRect();
        m_setting.szWindow.x = rc.right  - rc.left;
        m_setting.szWindow.y = rc.bottom - rc.top;

        // Create window
        m_hWnd = CreateWindowEx(m_setting.styleEx, g_nameClass, m_setting.title, m_setting.style, 
                                m_setting.ptWindow.x, m_setting.ptWindow.y, 
                                m_setting.szWindow.x, m_setting.szWindow.y, 
                                m_hParent, m_setting.hMenu, g_hInstance, null);

        return (m_hWnd != null);
    }


    bool destroy()
    {
        bool ret = cast(bool)DestroyWindow(m_hWnd);
        m_hWnd = null;
        return ret;
    }
    

    void invalidateRect(bool isClear=true)
    {
        InvalidateRect(m_hWnd, null, isClear);
    }


    void setCapture()
    {
        SetCapture(m_hWnd);
    }


    void releaseCapture()
    {
        ReleaseCapture();
    }


    void move(int x, int y, bool redraw=true)
    {
        MoveWindow(m_hWnd, x, y, m_setting.szWindow.x, m_setting.szWindow.y, redraw);
    }


    void move(int x, int y, int w, int h, bool redraw=true)
    {
        MoveWindow(m_hWnd, x, y, w, h, redraw);
    }


    void setTop()
    {
        SetWindowPos(m_hWnd, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
    }    


    void setTopmost()
    {
        SetWindowPos(m_hWnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
    }

    void toScreen(inout int x, inout int y)
    {
        POINT pt;
        pt.x = x;
        pt.y = y;
        ClientToScreen(m_hWnd, &pt);
        x = pt.x;
        y = pt.y;
    }

    void toClient(inout int x, inout int y)
    {
        POINT pt;
        pt.x = x;
        pt.y = y;
        ScreenToClient(m_hWnd, &pt);
        x = pt.x;
        y = pt.y;
    }
    
    bool isInClient(int x, int y)
    {
        POINT pt = getClientSize();
        return (x>=0 && x<pt.x && y>=0 && y<pt.y);
    }
    
    POINT getClientSize()
    {
        RECT rc;
        GetClientRect(m_hWnd, &rc);
        m_setting.szClient.x = rc.right;
        m_setting.szClient.y = rc.bottom;
        return m_setting.szClient;
    }
    
    
//----------------------------------------------------------------------
//  Operation of child window
//----------------------------------------------------------------------
protected:
    /// Call from proc() when it's necessary to handle message by child window.
    bool procChild(HWND hWnd, uint WM, WPARAM wParam, LPARAM lParam)
    {
        foreach(IWindow wnd; m_listChild) {
            if (wnd.proc(hWnd, WM, wParam, lParam)) return true;
        }
        return false;
    }
    
    
    /// Resister child wnd.
    void addChild(IWindow wnd)
    {
        m_listChild.push_back(wnd);
    }
    
        


//----------------------------------------------------------------------
//  Access to setting information
//----------------------------------------------------------------------
public:
    /// Window setting
    void setWindow(char[] title, int width, int height)
    {
        m_setting.title      = title;
        m_setting.szClient.x = width;
        m_setting.szClient.y = height;
    }
    
    
    /// Calculate window rect
    RECT calcWindowRect()
    {
        RECT rc;

        if (m_setting.szClient.x == 0 && m_setting.szClient.y == 0) {
            rc.left  = rc.top = 0;
            rc.right = rc.bottom = CW_USEDEFAULT;
        } else {
            rc.left  = rc.top = 0;
            rc.right = m_setting.szClient.x;
            rc.bottom = m_setting.szClient.y;
            AdjustWindowRectEx(&rc, m_setting.style, (m_setting.hMenu != null), m_setting.styleEx);
        }
        
        return rc;
    }

    
    
    
//---------------------------------------------------
//  Windows message parser
//---------------------------------------------------
public:
    static void parseMouseEvent(WPARAM wParam, LPARAM lParam, out int x, out int y, out int flag)
    {
        flag = wParam;
        x = cast(short)(lParam & 0xffff);
        y = cast(short)((lParam>>16) & 0xffff);
    }




//---------------------------------------------------
//  Virtual function
//---------------------------------------------------
public:
    /// Callc back from main loop. Retern false, when the work isnot completed.
    bool idle()
    {
        return true;
    }
    
    
    /// Callc back from window procedure.
    bool proc(HWND hWnd, uint WM, WPARAM wParam, LPARAM lParam)
    {
        return procChild(hWnd, WM, wParam, lParam);
    }
}



